@@ -11,7 +11,7 @@ Note the following:
     ALF files must always have an extension
 
 For more information, see the following documentation:
-    https://int-brain-lab.github.io/iblenv/one_docs/one_reference.html#alf
+    https://int-brain-lab.github.io/iblenv/one_docs/one_reference.html#alf  # FIXME Change link
 
 Created on Tue Sep 11 18:06:21 2018
 
@@ -157,11 +157,11 @@ def rel_path_parts(rel_path, as_dict=False, assert_valid=True):
         if any(m.values() if as_dict else m):
             return OrderedDict(**empty_collection, **m) if as_dict else (None, None, *m)
         else:
-            empty = re.compile(regex(f'{COLLECTION_SPEC}/{FILE_SPEC}')).groupindex.keys()
+            empty = re.compile(regex(f'{COLLECTION_SPEC}{FILE_SPEC}')).groupindex.keys()
             return OrderedDict.fromkeys(empty) if as_dict else tuple([None] * len(empty))
     else:
         folders, filname = par_path
-        rel_parts = re.match(regex(f'^{COLLECTION_SPEC}$'), folders)
+        rel_parts = re.match(regex(f'^{COLLECTION_SPEC}$'), folders + '/')
         file_parts = filename_parts(filname, as_dict=as_dict, assert_valid=assert_valid)
         if rel_parts:
             return OrderedDict(**rel_parts.groupdict(), **file_parts)\
@@ -183,7 +183,7 @@ def filename_parts(filename, as_dict=False, assert_valid=True):
     if m:  # py3.8
         return OrderedDict(m.groupdict()) if as_dict else m.groups()
     elif assert_valid:
-        raise ValueError('Invalid ALF filename')
+        raise ValueError(f'Invalid ALF filename: "{filename}"')
     else:
         return empty if as_dict else empty.values()
 
